/**
 * @license
 * Copyright 2022 Google LLC
 * SPDX-License-Identifier: Apache-2.0
 */
var __classPrivateFieldSet = (this && this.__classPrivateFieldSet) || function (receiver, state, value, kind, f) {
    if (kind === "m") throw new TypeError("Private method is not writable");
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a setter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot write private member to an object whose class did not declare it");
    return (kind === "a" ? f.call(receiver, value) : f ? f.value = value : state.set(receiver, value)), value;
};
var __classPrivateFieldGet = (this && this.__classPrivateFieldGet) || function (receiver, state, kind, f) {
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a getter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot read private member from an object whose class did not declare it");
    return kind === "m" ? f : kind === "a" ? f.call(receiver) : f ? f.value : state.get(receiver);
};
var _DiagnosticPrinter_instances, _DiagnosticPrinter_cwd, _DiagnosticPrinter_offsetConverterCache, _DiagnosticPrinter_printSupplemental, _DiagnosticPrinter_formatPath, _DiagnosticPrinter_offsetToPosition, _DiagnosticPrinter_getNewlineIndexes;
import * as pathlib from 'path';
/**
 * A known Wireit error.
 *
 * All errors that Wireit can anticipate should be an instance of this class.
 * Any other exception that is raised to the top-level should be considered a
 * bug.
 */
export class WireitError extends Error {
    /**
     * @param event The failure event that caused this exception.
     */
    constructor(event) {
        // Note that we need to pass some message for the base class, but it won't
        // usually be used. Most details are contained by the event, which can be
        // displayed nicely to the user by passing to a Logger instance.
        super(event.reason);
        Object.defineProperty(this, "event", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        this.event = event;
    }
}
export class DiagnosticPrinter {
    /**
     * @param workingDir Paths are printed relative to this directory.
     */
    constructor(workingDir) {
        _DiagnosticPrinter_instances.add(this);
        _DiagnosticPrinter_cwd.set(this, void 0);
        _DiagnosticPrinter_offsetConverterCache.set(this, new WeakMap());
        __classPrivateFieldSet(this, _DiagnosticPrinter_cwd, workingDir, "f");
    }
    print(diagnostic) {
        const path = __classPrivateFieldGet(this, _DiagnosticPrinter_instances, "m", _DiagnosticPrinter_formatPath).call(this, diagnostic.location);
        let result = `❌ ${path} ${diagnostic.message}
${drawSquiggle(diagnostic.location, 4)}`;
        if (diagnostic.supplementalLocations) {
            for (const supplementalLocation of diagnostic.supplementalLocations) {
                result += '\n\n' + __classPrivateFieldGet(this, _DiagnosticPrinter_instances, "m", _DiagnosticPrinter_printSupplemental).call(this, supplementalLocation);
            }
        }
        return result;
    }
}
_DiagnosticPrinter_cwd = new WeakMap(), _DiagnosticPrinter_offsetConverterCache = new WeakMap(), _DiagnosticPrinter_instances = new WeakSet(), _DiagnosticPrinter_printSupplemental = function _DiagnosticPrinter_printSupplemental(supplemental) {
    const squiggle = drawSquiggle(supplemental.location, 8);
    const path = __classPrivateFieldGet(this, _DiagnosticPrinter_instances, "m", _DiagnosticPrinter_formatPath).call(this, supplemental.location);
    return `    ${path} ${supplemental.message}\n${squiggle}`;
}, _DiagnosticPrinter_formatPath = function _DiagnosticPrinter_formatPath(location) {
    const relPath = pathlib.relative(__classPrivateFieldGet(this, _DiagnosticPrinter_cwd, "f"), location.file.path);
    const { line, column } = __classPrivateFieldGet(this, _DiagnosticPrinter_instances, "m", _DiagnosticPrinter_offsetToPosition).call(this, location.file, location.range.offset);
    return `${CYAN}${relPath}${RESET}:${YELLOW}${line}${RESET}:${YELLOW}${column}${RESET}`;
}, _DiagnosticPrinter_offsetToPosition = function _DiagnosticPrinter_offsetToPosition(file, offset) {
    const indexes = __classPrivateFieldGet(this, _DiagnosticPrinter_instances, "m", _DiagnosticPrinter_getNewlineIndexes).call(this, file);
    return indexes.toPosition(offset);
}, _DiagnosticPrinter_getNewlineIndexes = function _DiagnosticPrinter_getNewlineIndexes(file) {
    let converter = __classPrivateFieldGet(this, _DiagnosticPrinter_offsetConverterCache, "f").get(file);
    if (converter === undefined) {
        converter = new OffsetToPositionConverter(file.contents);
        __classPrivateFieldGet(this, _DiagnosticPrinter_offsetConverterCache, "f").set(file, converter);
    }
    return converter;
};
export class OffsetToPositionConverter {
    constructor(contents) {
        Object.defineProperty(this, "newlineIndexes", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        const indexes = [];
        for (let i = 0; i < contents.length; i++) {
            if (contents[i] === '\n') {
                indexes.push(i);
            }
        }
        this.newlineIndexes = indexes;
    }
    toPosition(offset) {
        if (this.newlineIndexes.length === 0) {
            return { line: 1, column: offset + 1 };
        }
        const line = this.newlineIndexes.findIndex((index) => index >= offset);
        if (line === 0) {
            return { line: 1, column: offset + 1 };
        }
        if (line === -1) {
            return {
                line: this.newlineIndexes.length + 1,
                column: offset - this.newlineIndexes[this.newlineIndexes.length - 1],
            };
        }
        return { line: line + 1, column: offset - this.newlineIndexes[line - 1] };
    }
}
const CYAN = '\x1b[36m';
const YELLOW = '\x1b[33m';
const RED = '\x1b[31m';
const RESET = '\x1b[0m';
// Exported for testing
export function drawSquiggle(location, indent) {
    let { range: { offset, length }, } = location;
    const fileContents = location.file.contents;
    const startOfInitialLine = fileContents.slice(0, offset).lastIndexOf('\n') + 1;
    const uncorrectedFirstNewlineIndexAfter = fileContents
        .slice(offset + length)
        .indexOf('\n');
    const endOfLastLine = uncorrectedFirstNewlineIndexAfter === -1
        ? undefined
        : offset + length + uncorrectedFirstNewlineIndexAfter;
    offset = offset - startOfInitialLine;
    const sectionToPrint = fileContents.slice(startOfInitialLine, endOfLastLine);
    let result = '';
    for (const line of sectionToPrint.split('\n')) {
        result += `${' '.repeat(indent)}${line}\n`;
        const squiggleLength = Math.min(line.length - offset, length);
        result +=
            ' '.repeat(offset + indent) +
                `${RED}${'~'.repeat(squiggleLength)}${RESET}\n`;
        offset = 0;
        length -= squiggleLength + 1; // +1 to account for the newline
    }
    // Drop the last newline.
    return result.slice(0, -1);
}
//# sourceMappingURL=error.js.map